#include "btio.h"
#include "bte.h"

/* System general includes */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/time.h>
#include <errno.h>


static int portfd = 0;
static int portflags = 0;
static int oldflags = 0;
static struct termios oldterm;
static struct termios portterm;
static int rtscts=0;
static int get_baudrate(int);

int bt_init_port(char* port) {
  int baudrate = get_baudrate(btecfg.baudrate);

  /* get a file descriptor */
  if ((portfd = open(port, O_RDWR | O_NOCTTY | O_NDELAY)) < 0)
    {
      portfd = 0;
      return -1;
    }
  
  /* check to see that the file is a terminal */
  if (!isatty(portfd))
    {
      return -1;
    }
  
  /* get port attributes, store in oldterm */
  if (tcgetattr(portfd, &oldterm) < 0)
    {
      return -1;
    }

  /* get port flags, save in oldflags */
  if ((oldflags = fcntl(portfd, F_GETFL)) < 0)
    {
      return -1;
    }

  portterm = oldterm;
  portflags = oldflags;

  portterm.c_cflag = baudrate | CRTSCTS | CS8 | CLOCAL | CREAD | O_NDELAY;
  if (!rtscts)
    portterm.c_cflag &= ~CRTSCTS;
  portterm.c_iflag = IGNPAR;
  portterm.c_oflag = 0;
  portterm.c_lflag = 0;
  portterm.c_cc[VTIME]    = 0;
  portterm.c_cc[VMIN]     = 0;
  tcflush(portfd, TCIOFLUSH);
  if (tcsetattr(portfd,TCSANOW,&portterm) < 0) 
    {
      return -1;
    }
  /* set non-blocking */
  if (fcntl(portfd, F_SETFL, (portflags |= O_NONBLOCK)) < 0)
    {
      return -1;
    }
  return 0;
}

int bt_open_port(char* port) {
  return bt_init_port(port);
}

int bt_close_port() {
  int retval = 0;

  if (!portfd)
    {			/* already closed */
      return -1;
    }
  
  /* restore old settings */
  if (tcsetattr(portfd, TCSADRAIN, &oldterm) < 0)
    {
      retval = -1;
    }
  
  if (fcntl(portfd, F_SETFL, oldflags) < 0)
    {
      retval = -1;
    }
  
  close(portfd);
  portfd = 0;
  
  return retval;
}

int bt_read(char* buf, int len) {
  return read(portfd, buf, len);
}

int bt_read_tmo(char* buf, int len, int timeout) {
  int ok;
  fd_set rdfds;
  struct timeval tv;
  
  FD_ZERO(&rdfds);
  FD_SET(portfd, &rdfds);

  /* block until something to read or timeout occurs.  select() is damn cool */
  
  if (timeout < 0)
    {
      ok = select(portfd + 1, &rdfds, NULL, NULL, NULL);
    }
  else
    {
      tv.tv_sec = timeout / 1000000;
      tv.tv_usec = (timeout % 1000000);
      ok = select(portfd + 1, &rdfds, NULL, NULL, &tv);
    }
  
  if (ok > 0)
    {
      ok = read(portfd, buf, len);
      if (ok == 0)
	{
	  return EOF;
	}
      return ok;

    } else {
      if (ok < 0)
	{
	  return -1;
	}
      else
	{
	  errno = ETIMEDOUT;
	  return 0;
	}
    }
}

int  bt_readchar(int timeout) {
  char buf[1];
  int rv =  bt_read_tmo(buf, 1, timeout);

  if (rv == 1) return buf[0];

  if (rv == 0) return EOF - 1;
  
  return EOF;
}

int bt_writechar(char data) {
  char buf[1];
  int rv = bt_write(buf, 1);
  return (rv == 1 ? 0 : -1);
}  

int bt_write(char* buf, int len) {
  return write(portfd, buf, len);
}

int bt_getfd(void) {
  return portfd;
}
#define WINDOWS 1

int bt_put_command(char* command,
		   char* answer,
		   int   max,
		   int   timeout,
		   char* expect) {
  int count=0;
  int readcount;
  int toread;
  char tmp[100];
  int timeoutcounter=0;
  int found=0;
  int available;
  int status;
  char loganswer[2048];

  // send command
  if (command && command[0])
  {
  // Cycwin does not support TIOC functions and has not workaround for this.
  // So do not check CTS on Cygwin.
#ifndef WINDOWS
    if (rtscts)
    {
      ioctl(portfd, TIOCMGET, &status);
      while (!(status & TIOCM_CTS))
      {
        usleep(100000);
        timeoutcounter++;
        ioctl(portfd, TIOCMGET, &status);
        if (timeoutcounter>=timeout)
        {
          return 0;
        }
      }
    }
#endif
    write(portfd,command,strlen(command));
    tcdrain(portfd);
  }


  answer[0]=0;
  do
    {
      // try to read some bytes.
      
      // Cygwin does not support TIOC functions and has not workaround for this.
      // So do not check number of available bytes
#ifndef WINDOWS
      ioctl(portfd,FIONREAD,&available);	// how many bytes are available to read?
      if (available<1)			// if 0 then wait a little bit and retry
	{
	  usleep(100000);
	  //write(1,".",1);
	  timeoutcounter++;
	  ioctl(portfd,FIONREAD,&available);
	}
#else
      // Only for Windows. Read as much as possible
      usleep(100000);
      //write(1,".",1);
      timeoutcounter++;
      available=sizeof(tmp)-1;
#endif
      if (available>0)
	{
	  // How many bytes do I want to read maximum? Not more than tmp buffer size.
	  toread=max-count-1;
	  if (toread>sizeof(tmp)-1)
	    toread=sizeof(tmp)-1;
	  // And how many bytes are available?
	  if (available<toread)
	    toread=available;
	  // read data
	  readcount=read(portfd,tmp,toread);
	  if (readcount<0)
	    readcount=0;
	  tmp[readcount]=0;
	  // add read bytes to the output buffer
	  if (readcount)
	    {
	      strcat(answer,tmp);
	      count+=readcount;
	      
	      // if we have more time to read, check if we got already the expected string
	      if ((timeoutcounter<timeout) && (found==0))
		{
		  // check if it's the expected answer
		  if ((strstr(answer,"OK\r")) || (strstr(answer,"ERROR")))
		    found=1;
		  if (expect) if (expect[0])
		    if (strstr(answer,expect))
		      found=1;
		  // if found then set timoutcounter to read only 0.1s after that and not more
		  if (found)
		    timeoutcounter=timeout-1;
		}
	    }
	}
    }
  // repeat until timout
  while (timeoutcounter<timeout);

  return count;
  
}


static int get_baudrate(int rate) {
  int baudrate;

  switch (rate) {
    case 300:
      baudrate = B300;
      break;
    case 1200:
      baudrate = B1200;
      break;
    case 2400:
      baudrate = B2400;
      break;
    case 9600:
      baudrate = B9600;
      break;
    case 19200:
      baudrate = B19200;
      break;
    case 38400:
      baudrate = B38400;
      break;
#ifdef B57600
    case 57600:
      baudrate = B57600;
      break;
#endif
#ifdef B115200
  case 115200:
    baudrate = B115200;
    break;
#endif
#ifdef B230400
  case 230400:
    baudrate = B230400;
    break;
#endif
  default:
    printf("bad baudrate %d, defaulting to 9600\n", rate);
    baudrate = B9600;
  }

  return baudrate;
  
}
